from src import get_cache_size
from src.primesgen import Eratosthenes, IndexedEratosthenes

segment_size = get_cache_size()


def test_100_000(): assert_len_at(100_000)


def test_1_000_000(): assert_len_at(1_000_000)


def test_10_000_000(): assert_len_at(10_000_000)


def test_100_000_000(): assert_len_at(100_000_000)


def test_35_123_231(): assert_len_at(35_123_231)


def assert_len_at(max_val):
    c_primes = Eratosthenes().primes(max_val)
    c_len = len(c_primes)
    c_primes.clear()

    i_primes = IndexedEratosthenes().primes(max_val)
    i_len = len(i_primes)
    i_primes.clear()

    assert c_len == i_len, f'{max_val}, lengths differ ({c_len}, {i_len})'



import inspect
import platform
import re
import subprocess

sieve_factory_erat_indexed_cutoff = 800_000_000
on_disk_multiples_block_size = 25_000
multiples_factory_ondisk_inmem_cutoff = 1_000


def get_cache_size() -> int:
    os = platform.system()
    match os:
        case 'Darwin':
            return _get_darwin_cache_size()
        case 'Linux':
            # not tested, may raise
            return _untested_get_linux_cache_size()
        case 'Windows':
            raise NotImplementedError(f'not implemented for your operating system: {os}')
        case _:
            raise NotImplementedError(f'not implemented for your operating system: {os}')


def _untested_get_linux_cache_size() -> int | None:
    try:
        # Run the lscpu command and capture its output
        output = subprocess.run(['lscpu'], capture_output=True, text=True, check=True).stdout
        # The output format for L1d cache size typically looks like: "L1d cache:             32 KiB"
        matches = re.search(r'L1d cache:\s+([0-9.]+)\s+([KMGT]?iB)', output)
        if matches is not None:
            size_val = float(matches.group(1))
            exponent = 0
            match matches.group(2):
                # Convert to bytes for a consistent value
                case 'KiB':
                    exponent = 1
                case 'MiB':
                    exponent = 2
                case 'GiB':
                    exponent = 3
                case _:
                    pass  # already set to 0
            return int(size_val * 1024 ** exponent)
    except Exception as e:
        e.add_note(
            f'not implemented correctly for Linux\nhere is the offending code in __init__.py\n{"".join(inspect.getsourcelines(_get_linux_cache_size)[0])}')
        raise e


def _get_darwin_cache_size() -> int:
    raw_str = subprocess.run(['sysctl', 'hw.l1dcachesize'], capture_output=True, text=True, check=True).stdout
    return int(raw_str.strip().split(' ')[-1])



from abc import ABC, abstractmethod


def factory(in_queue=None) -> Primes:
    return PrimesToReturn() if in_queue is None else PrimesToQueue(in_queue)


class Primes(ABC):

    @abstractmethod
    def append(self, prime: int) -> None: pass

    @abstractmethod
    def all(self) -> list[int]: pass


class PrimesToReturn(Primes):
    """
    uses a local array
    """

    def __init__(self) -> None:
        self._primes = []

    def append(self, prime: int) -> None: self._primes.append(prime)

    def all(self) -> list[int]: return self._primes


class PrimesToQueue(Primes):
    """
    uses a queue
    """

    def __init__(self, in_queue) -> None:
        self._queue = in_queue
        self._queue_count = 0

    def append(self, prime: int) -> None:
        self._queue.put(prime)
        self._queue_count += 1

    def all(self) -> list[int]:
        self._queue.put(None)
        return []



import datetime
import json
import linecache
import math
from abc import ABC, abstractmethod
from pathlib import Path
from typing import Iterable

from src import on_disk_multiples_block_size, multiples_factory_ondisk_inmem_cutoff


def factory(max_possible_value) -> Multiples:
    return InMemMultiples(
        max_possible_value) if max_possible_value < multiples_factory_ondisk_inmem_cutoff else OnDiskMultiples(
        max_possible_value, on_disk_multiples_block_size)


class Multiple:

    def __init__(self, start_val: int, p_val: int, max_val: int = None, advance: bool = True) -> None:
        self.start_val = start_val
        self.p_val = int(p_val)
        if advance:
            self.advance(max_val)

    def advance(self, max_val: int) -> None:
        self.start_val = self._get_start_val_in_next_seg(self.start_val, self.p_val, max_val)

    @staticmethod
    def _get_start_val_in_next_seg(start_val: int, p_val: int, max_val: int) -> int:
        max_val_plus_1_seg = max_val + 1
        # change anything here to make life hell
        return int(start_val + (-max_val_plus_1_seg if start_val > max_val else (math.floor(
            ((max_val - start_val) / p_val)) + 1) * p_val - max_val_plus_1_seg))


class Multiples(ABC):

    @abstractmethod
    def add(self, multiple: Multiple) -> None: pass

    @abstractmethod
    def advance(self, max_val: int) -> None: pass

    @abstractmethod
    def __iter__(self) -> Iterable[Multiple]: pass


class InMemMultiples(Multiples):
    def __init__(self, max_possible_value) -> None:
        self.max_possible_value = max_possible_value
        self._multiples = []

    def __len__(self) -> int:
        return self._multiples.__len__()

    def __iter__(self) -> Iterable[Multiple]:
        return self._multiples.__iter__()

    def add(self, multiple: Multiple) -> None:
        if multiple.start_val < self.max_possible_value:
            self._multiples.append(multiple)

    def advance(self, max_val: int) -> None:
        for i in range(len(self._multiples)):
            self._multiples[i].advance(max_val)


class OnDiskMultiples(Multiples):

    def __init__(self, max_possible_value: int, max_mem_buffer: int, tmp_dir: str = '/var/tmp') -> None:
        self._max_possible_value = max_possible_value
        self._max_mem_buffer = max_mem_buffer
        self._tempfile = f'{tmp_dir}/{datetime.datetime.now().strftime("primesgenseg_multipes_%Y-%m-%d_%H-%M-%S")}.dat'
        Path(self._tempfile).touch()
        self._block_number_loaded = 1
        self.__in_mem_data = []

    def __len__(self) -> int:
        return len(self.__in_mem_data) + self._max_mem_buffer * (self._block_number_loaded - 1)

    def __iter__(self) -> Iterable[Multiple]:
        yield from self.__in_mem_data
        for block_num in range(1, self._block_number_loaded):
            yield from self._load_block(block_num)

    def advance(self, max_val: int) -> None:
        for m in self.__in_mem_data:
            m.advance(max_val)
        for block_num in range(1, self._block_number_loaded):
            block_data = self._load_block(block_num)
            for i in range(len(block_data)):
                block_data[i].advance(max_val)
            self._write_block(block_num, block_data)

    def add(self, multiple: Multiple) -> None:
        if multiple.start_val <= self._max_possible_value:
            self.__in_mem_data.append(multiple)

        if len(self.__in_mem_data) > self._max_mem_buffer - 1:
            self._write_block(self._block_number_loaded, self.__in_mem_data)
            self.__in_mem_data.clear()
            self._block_number_loaded += 1

    def _write_block(self, block_num, block_data: list[Multiple]) -> None:
        if len(block_data) != 0:
            to_write = json.dumps({block_num: [self.to_dict(d) for d in block_data]})
            with open(self._tempfile, 'r') as fd:
                lines = fd.readlines()
            with open(self._tempfile, 'w+') as fd:
                if len(lines) < block_num:
                    lines.append(to_write)
                elif len(lines) == block_num:
                    lines[block_num - 1] = to_write
                out = '\n'.join((line.strip('\n') for line in lines if len(line) > 0))
                fd.writelines(f'{out}')
                self._data_on_disk = True

    def _load_block(self, block_num: int) -> list[Multiple]:
        return [self.from_dict(m) for m in json.loads(linecache.getline(self._tempfile, block_num).strip('\n'))[
            f'{block_num}']] if self._data_on_disk and block_num != self._block_number_loaded else []

    @staticmethod
    def to_dict(multiple: Multiple) -> dict:
        return {"start_val": multiple.start_val, 'p_val': multiple.p_val}

    @staticmethod
    def from_dict(in_data: dict) -> Multiple:
        return Multiple(in_data['start_val'], in_data['p_val'], advance=False)



from src import get_cache_size
from src.multiples import factory as multiples_factory
from src.primes import factory as primes_factory
from src.primesgen import Eratosthenes
from src.primesgenseg import SegmentedEratosthenes

segment_size = get_cache_size()


def test_84_max_val_9_segment_size(): assert_len_at(84, 9)


def test_130_max_val_9_segment_size(): assert_len_at(130, 9)


def test_segment_boundaries():
    seg_sizes = [int(.1 * segment_size), int(.25 * segment_size), int(.5 * segment_size), int(.75 * segment_size),
                 segment_size, segment_size * 2]
    for seg_size in seg_sizes:
        assert_len_at(seg_size - 1, seg_size)
    for i in [-1, 0, 1, 2]:
        for seg_size in seg_sizes:
            assert_len_at(i, seg_size)
    for seg_size in seg_sizes:
        for m in range(2, 5):
            assert_len_at((m - 1) * seg_size, seg_size)
            assert_len_at((m - 1) * seg_size + 1, seg_size)


def test_100_000(): assert_len_at(100_000, segment_size)


def test_120_001(): assert_len_at(120_001, segment_size)


def assert_segment(segment_num):
    for i in range((segment_num - 1) * segment_size, segment_num * segment_size):
        assert_len_at(i, segment_size)


def assert_len_at(max_val, seg_size):
    c_primes = Eratosthenes().primes(max_val)
    c_len = len(c_primes)
    c_primes.clear()

    s_primes = SegmentedEratosthenes(seg_size, primes_factory(None), multiples_factory(max_val)).primes(max_val)
    s_len = len(s_primes)
    s_primes.clear()

    assert s_len == c_len, f'{max_val}, lengths differ ({s_len}, {c_len})'



import math
import time
from typing import Tuple

import numpy as np

from src import get_cache_size
from src.multiples import Multiples, Multiple, factory
from src.multiples import factory as multiples_factory
from src.primes import Primes
from src.primes import factory as primes_factory
from src.primesgen import Sieve


class SegmentedEratosthenes(Sieve):
    """
    Utilizes segmentation to handle numbers beyond the scope of the normal erat by reducing the mem used for the sieve

    The prime values can be broadcast, if a queue is provided, thus reducing the primes in mem
    please see the factory function in primes

    The multiples can be stored in mem, or offloaded to disk as needed
    please see the factory function in multiples

    Essentially a copy of Kim Walisch's C++ impl (which is way way way faster)
    https://github.com/kimwalisch/primesieve/wiki/Segmented-sieve-of-Eratosthenes

    but the off loading of the data to queues and/or disk is something I want to investigate more
    """

    def __init__(self, in_segment_max_size, in_primes: Primes, in_multiples: Multiples) -> None:
        self._seg_max_size = in_segment_max_size
        self._multiples = in_multiples
        self._primes = in_primes
        self._max_sqrt_val = None
        self._max_val = None

    def _create_sieve(self, seg_max_val: int, seg_offset: int, seg_num: int) -> np.ndarray[Tuple[int]]:
        new_sieve = np.ones(seg_max_val + 1, dtype=bool)
        # bye bye evens
        new_sieve[seg_offset % 2::2] = False
        if seg_num == 1:
            new_sieve[1] = False
        for m in self._multiples:
            new_sieve[m.start_val::m.p_val] = False
        self._multiples.advance(seg_max_val)
        return new_sieve

    def primes(self, in_max_val) -> list[int]:
        self._multiples = factory(in_max_val)
        if in_max_val < 3:
            return []
        # prep data for all segments based on input
        self._max_val = in_max_val
        self._max_sqrt_val = int(math.sqrt(self._max_val)) + 1
        num_segs = int((self._max_val + 1) / self._seg_max_size) + 1
        if self._seg_max_size == self._max_val + 1:
            num_segs = 1

        # sieve each segment
        for seg_num in range(1, num_segs + 1):
            self._sieve_segment(seg_num)
        return self._primes.all()

    def _sieve_segment(self, seg_num: int) -> None:
        seg_offset = (seg_num - 1) * self._seg_max_size
        max_val_seg = min(self._max_val - seg_offset, self._seg_max_size - 1)
        max_sqrt_val_seg = min(max(0, self._max_sqrt_val - seg_offset), self._seg_max_size)
        sieve = self._create_sieve(max_val_seg, seg_offset, seg_num)

        # erat (0 --> max_sqrt_val_seg)
        for p_val in (true_index + seg_offset for true_index in range(0, max_sqrt_val_seg) if sieve[true_index]):
            self._primes.append(p_val)
            p_2_val = p_val * p_val
            p_2_val_seg = p_2_val - seg_offset
            self._multiples.add(Multiple(p_2_val_seg, p_val, max_val_seg))
            sieve[p_2_val_seg::p_val] = False

        # rest (max_sqrt_val_seg -- > end          and        additional offset of max_sqrt_val_seg)
        for p_val in (true_index + max_sqrt_val_seg + seg_offset for true_index in
                      np.where(sieve[max_sqrt_val_seg::])[0]):
            self._primes.append(p_val)
            p_2_val = p_val * p_val
            self._multiples.add(Multiple(p_2_val - seg_offset, p_val, max_val_seg))


if __name__ == '__main__':
    max_val = 1_000_000_000
    # print(f'\torg found all {len(Eratosthenes().primes(max_val)):,} primes <= {max_val:,}')
    # print(f'\tidx found all {len(IndexedEratosthenes().primes(max_val)):,} primes <= {max_val:,}')
    primes = primes_factory()  # if we supply a queue, primes will be broadcast instead of returning vals (and doesn't store them after they are sent)
    multiples = multiples_factory(max_val)
    segment_max_size = get_cache_size()
    start = time.time()
    prime_values = SegmentedEratosthenes(segment_max_size, primes, multiples).primes(max_val)
    duration = time.time() - start
    print(f'\tseg found all {len(prime_values):,} primes <= {max_val:,} in {duration:.2f} seconds')
